#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from . import scoring, config, __version__
from .evolution_chamber import (
    CDSEvolutionChamber, IterationOptions, ExecutionOptions)
from .presets import load_preset
from .reporting import ReportGenerator
from Bio import SeqIO
import argparse
import shlex
import time
import sys
import os

SPECIES_ALIASES = {
    'human': 'Homo sapiens',
    'mouse': 'Mus musculus',
    'zebrafish': 'Danio rerio',
    'rat': 'Rattus norvegicus',
    'macaque': 'Macaca mulatta',
}

CONSERVATIVE_START_DEFAULT_WIDTH = 7
BOOST_LOOP_MUTATIONS_DEFAULT_WIDTH = 15

def preparse_config_preset_addons():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--preset', type=str, required=False, default=None)
    parser.add_argument('--addon', type=str, action='append')
    args, _ = parser.parse_known_args()

    preset = config.load_config()

    if args.preset is not None:
        try:
            preset.update(load_preset(open(args.preset).read()))
        except Exception:
            print(f'Failed to load the preset from {args.preset}.')
            sys.exit(1)

    addon_paths = []
    if preset is not None and 'addons' in preset:
        addon_paths.extend(preset['addons'])

    if args.addon is not None:
        for path in args.addon:
            if os.path.exists(path):
                addon_paths.append(path)
            else:
                print(f'Addon path {path} is missing.')
                sys.exit(1)

    return preset, addon_paths

def apply_preset(main_parser, preset):
    optmap = main_parser._option_string_actions

    def fix_option(opt, newval):
        if opt.default == newval:
            return

        opt.default = newval

        # Fix store_action for boolean options
        if isinstance(newval, bool):
            opt.const = not newval

        # Fix default value in help string
        if '(default:' in opt.help:
            prefix = opt.help.split('(default:')[0]
            opt.help = f'{prefix}(default: {newval})'

    ignore_options = ['addons', 'command_line']

    for argname, argval in preset.items():
        if argname in ignore_options:
            continue
        elif argname != 'fitness':
            optname = '--' + argname.replace('_', '-')
            fix_option(optmap[optname], argval)
            continue

        for grpname, grpvalues in argval.items():
            for optname, optval in grpvalues.items():
                optname = f'--{grpname}-{optname}'.replace('_', '-')
                fix_option(optmap[optname], optval)

def check_lineardesign(args):
    if args.lineardesign is None:
        return

    if args.lineardesign_dir is None:
        print('Specify the path to the LinearDesign top directory with '
              '--lineardesign-dir.', file=sys.stderr)
        sys.exit(1)

    executable = os.path.join(args.lineardesign_dir, 'bin', 'LinearDesign_2D')
    if not os.path.exists(executable):
        print('LinearDesign is not available. Please specify the correct path '
              'to the LinearDesign top directory with --lineardesign-dir.',
              file=sys.stderr)
        sys.exit(1)

def check_argument_validity(args):
    check_lineardesign(args)

    if args.conservative_start is not None:
        try:
            if args.conservative_start.count(':') == 1:
                cons_iter, cons_width = args.conservative_start.split(':', 1)
                cons_iter, cons_width = int(cons_iter), int(cons_width)
            elif ':' in args.conservative_start:
                raise ValueError
            else:
                cons_iter = int(args.conservative_start)
                cons_width = CONSERVATIVE_START_DEFAULT_WIDTH

            if not (1 <= cons_iter <= args.iterations):
                print('Invalid value for --conservative-start. ITER must be '
                      'between 1 and the maximum iteration number.',
                      file=sys.stderr)
                sys.exit(1)
            if cons_width <= 0:
                print('Invalid value for --conservative-start. WIDTH must be '
                      'a positive integer.', file=sys.stderr)
                sys.exit(1)
        except ValueError:
            print('Invalid format for --conservative-start. Use '
                'ITER[:WIDTH] format.', file=sys.stderr)
            sys.exit(1)
        else:
            args.conservative_start = f'{cons_iter}:{cons_width}'

    if args.boost_loop_mutations is not None:
        try:
            if args.boost_loop_mutations.count(':') == 1:
                boost_weight, boost_start = args.boost_loop_mutations.split(':', 1)
                boost_weight, boost_start = float(boost_weight), int(boost_start)
            elif ':' in args.boost_loop_mutations:
                raise ValueError
            else:
                boost_weight = float(args.boost_loop_mutations)
                boost_start = BOOST_LOOP_MUTATIONS_DEFAULT_WIDTH

            if boost_weight < 0:
                print('Invalid value for --boost-loop-mutations. WEIGHT must be '
                      'a non-negative number.', file=sys.stderr)
                sys.exit(1)
            if boost_start < 0:
                print('Invalid value for --boost-loop-mutations. START must be '
                      'a non-negative integer.', file=sys.stderr)
                sys.exit(1)
        except ValueError:
            print('Invalid format for --boost-loop-mutations. Use '
                'WEIGHT[:START] format.', file=sys.stderr)
            sys.exit(1)
        else:
            args.boost_loop_mutations = f'{boost_weight}:{boost_start}'

def parse_options(scoring_funcs, preset):
    parser = argparse.ArgumentParser(
        prog='vaxpress',
        description='VaxPress: A Codon Optimizer for mRNA Vaccine Design')

    grp = parser.add_argument_group('Input/Output Options')
    grp.add_argument('-i', '--input', required=True, metavar='FILE',
                     help='input fasta file containing the CDS sequence')
    grp.add_argument('--protein', default=False, action='store_true',
                     help='input is a protein sequence')
    grp.add_argument('-o', '--output', required=True, metavar='DIR',
                     help='output directory')
    grp.add_argument('--overwrite', action='store_true',
                     help='overwrite output directory if it already exists')
    grp.add_argument('-q', '--quiet', default=False, action='store_true',
                     help='do not print progress')
    grp.add_argument('--print-top', type=int, default=10, metavar='N',
                     help='print top and bottom N sequences (default: 10)')
    grp.add_argument('--report-interval', type=int, default=5, metavar='MIN',
                     help='report interval in minutes (default: 5)')
    grp.add_argument('--version', action='version', version=__version__)

    grp = parser.add_argument_group('Execution Options')
    grp.add_argument('--preset', type=str, required=False, default=None,
                     metavar='FILE', help='use preset values in parameters.json')
    grp.add_argument('--addon', type=str, action='append', metavar='FILE',
                     help='load a third-party fitness function')
    grp.add_argument('-p', '--processes', type=int, default=4, metavar='N',
                     help='number of processes to use (default: 4)')
    grp.add_argument('--seed', type=int, default=922, metavar='NUMBER',
                     help='random seed (default: 922)')
    grp.add_argument('--folding-engine', default='vienna', metavar='NAME',
                     choices=['vienna', 'linearfold'],
                     help='RNA folding engine: vienna or linearfold '
                          '(default: vienna)')

    grp = parser.add_argument_group('Optimization Options')
    grp.add_argument('--random-initialization', action='store_true',
                     default=False, help='randomize all codons at the beginning')
    grp.add_argument('--conservative-start', default=None, metavar='ITER[:WIDTH]',
                     help='conserve sequence for the first ITER iterations '
                          'except the first WIDTH amino acids')
    grp.add_argument('--iterations', type=int, default=10, metavar='N',
                     help='number of iterations (default: 10)')
    grp.add_argument('--offsprings', type=int, default=20, metavar='N',
                     help='number of offsprings per iteration (default: 20)')
    grp.add_argument('--survivors', type=int, default=2, metavar='N',
                     help='number of survivors per iteration (default: 2)')
    grp.add_argument('--initial-mutation-rate', type=float, default=0.1,
                     metavar='RATE',
                     help='initial mutation rate (default: 0.1)')
    grp.add_argument('--boost-loop-mutations',
                     default=f'3:{BOOST_LOOP_MUTATIONS_DEFAULT_WIDTH}',
                     metavar='WEIGHT[:START]', type=str,
                     help='boost mutations in loops after position START '
                          'by WEIGHT (default: 3:'
                          f'{BOOST_LOOP_MUTATIONS_DEFAULT_WIDTH})')
    grp.add_argument('--winddown-trigger', type=int, default=15, metavar='N',
                     help='number of iterations with the same best score to '
                          'trigger mutation stabilization (default: 15)')
    grp.add_argument('--winddown-rate', type=float, default=0.9, metavar='RATE',
                     help='mutation rate multiplier when mutation stabilization '
                          'is triggered (default: 0.9)')
    grp.add_argument('--species', default='human', metavar='NAME',
                     help='target species (default: human)')
    grp.add_argument('--codon-table', default='standard', metavar='NAME',
                     help='codon table (default: standard)')

    grp = parser.add_argument_group('LinearDesign Options')
    grp.add_argument('--lineardesign', type=float, default=None,
                     metavar='LAMBDA',
                     help='call LinearDesign to initialize the optimization')
    grp.add_argument('--lineardesign-dir', type=str, default=None, metavar='DIR',
                     help='path to the top directory containing LinearDesign')
    grp.add_argument('--lineardesign-omit-start', type=int, metavar='AA', default=5,
                     help='number of amino acids to omit from the N-terminus '
                          'when calling LinearDesign (default: 5)')

    argmaps = []
    for func in sorted(scoring_funcs.values(), key=lambda f: f.priority):
        argmap = func.add_argument_parser(parser)
        argmaps.append((func, argmap))

    if preset:
        apply_preset(parser, preset)

    args = parser.parse_args()
    scoring_opts = {}
    for func, argmap in argmaps:
        opts = scoring_opts[func.name] = {}
        for optname, varname in argmap:
            opts[varname] = getattr(args, optname[2:].replace('-', '_'))

    config.initialize_config_if_needed(args)
    check_argument_validity(args)

    return args, scoring_opts

def run_vaxpress():
    preset, addon_paths = preparse_config_preset_addons()
    scoring_funcs = scoring.discover_scoring_functions(addon_paths)

    args, scoring_options = parse_options(scoring_funcs, preset)

    inputseq = next(SeqIO.parse(args.input, 'fasta'))
    seqdescr = inputseq.description
    cdsseq = str(inputseq.seq)

    command_line = ' '.join(shlex.quote(arg) for arg in sys.argv)

    iteration_options = IterationOptions(
        n_iterations=args.iterations,
        n_offsprings=args.offsprings,
        n_survivors=args.survivors,
        initial_mutation_rate=args.initial_mutation_rate,
        winddown_trigger=args.winddown_trigger,
        winddown_rate=args.winddown_rate,
    )

    execution_options = ExecutionOptions(
        output=args.output,
        command_line=command_line,
        overwrite=args.overwrite,
        seed=args.seed,
        processes=args.processes,
        random_initialization=args.random_initialization,
        conservative_start=args.conservative_start,
        boost_loop_mutations=args.boost_loop_mutations,
        species=SPECIES_ALIASES.get(args.species, args.species),
        codon_table=args.codon_table,
        quiet=args.quiet,
        seq_description=seqdescr,
        print_top_mutants=args.print_top,
        protein=args.protein,
        addons=addon_paths,
        lineardesign_dir=args.lineardesign_dir,
        lineardesign_lambda=args.lineardesign,
        lineardesign_omit_start=args.lineardesign_omit_start,
        folding_engine=args.folding_engine,
    )

    next_report = 0 # Generate the first report immediately.
    # vaxpress assumes that the system clock does not go back or jump forward.

    try:
        evochamber = CDSEvolutionChamber(
            cdsseq, scoring_funcs, scoring_options,
            iteration_options, execution_options)

        status = None
        for status in evochamber.run():
            do_report = len(status['time']) > 1 and (
                (status['iter_no'] < 0) or (status['time'][-1] >= next_report))

            if do_report:
                next_report = status['time'][-1] + args.report_interval * 60
                if status['iter_no'] >= 0:
                    evochamber.printmsg('==> Generating intermediate report...')

                evaldata = evochamber.save_results()
                status.update({'evaluations': evaldata, 'version': __version__})

                generate_report(status, args, scoring_options, iteration_options,
                                execution_options, inputseq, scoring_funcs)

        finished = (status is not None and status['iter_no'] < 0
                    and status['error'] == 0)
        if finished:
            evochamber.printmsg(
                'Finished successfully. You can view the results '
                f'in {evochamber.outputdir.rstrip("/")}/report.html.')

        return status['error'] if status is not None else 1
    except KeyboardInterrupt:
        return 1
    except FileExistsError:
        print('Output directory already exists. Use --overwrite '
              'option to overwrite it.')
        return 1

def generate_report(status, args, scoring_options, iteration_options,
                    execution_options, inputseq, scoring_funcs):
    if status['iter_no'] > 0: # Intermediate report
        total_elapsed = status['time'][-1] - status['time'][0]
        time_per_iteration = total_elapsed / status['iter_no']
        remaining = (iteration_options.n_iterations -
                     status['iter_no']) * time_per_iteration

        expected_end = time.time() + remaining

        status['speed'] = time_per_iteration
        status['expected_end'] = expected_end
        status['progress_percentage'] = int(100 * status['iter_no'] /
                                            iteration_options.n_iterations)
        status['refresh'] = args.report_interval * 60 + 5 # 5 seconds for safety

    ReportGenerator(status, args, scoring_options, iteration_options,
                    execution_options, inputseq, scoring_funcs).generate()

if __name__ == '__main__':
    ret = run_vaxpress()
    sys.exit(ret)



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import os
import subprocess as sp

def run_lineardesign(lineardesign_dir, sequence, lmd=0.5,
                     codonusage='codon_usage_freq_table_human.csv'):
    lineardesign_bin = os.path.join(lineardesign_dir, 'bin/LinearDesign_2D')
    lineardesign_bin = os.path.abspath(lineardesign_bin)
    if not os.path.exists(lineardesign_bin):
        raise FileNotFoundError(f'LinearDesign binary not found in {lineardesign_dir}')

    with sp.Popen([lineardesign_bin, str(lmd), '0', codonusage],
                  cwd=lineardesign_dir, stdin=sp.PIPE, stdout=sp.PIPE) as proc:
        proc.stdin.write(sequence.encode())
        proc.stdin.write(b'\n')
        proc.stdin.close()

        ret = proc.stdout.read().decode()

    if 'mRNA sequence:  ' not in ret:
        raise ValueError('Unexpected output from LinearDesign.')

    lines = ret.split('mRNA sequence:  ')[1].splitlines()
    rnaseq = lines[0].strip()
    rnastr = lines[1].split(': ')[1].strip()
    mfe = float(lines[2].split(': ')[1].split()[0])
    cai = float(lines[2].split(': ')[2].strip())

    return {
        'seq': rnaseq,
        'str': rnastr,
        'mfe': mfe,
        'cai': cai,
    }



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from . import ScoringFunction
import numpy as np

def gc_content_sliding_window(seq, winsize, stride):
    chars = np.frombuffer(seq.encode(), dtype=np.uint8)
    isgc = ((chars == ord('G')) + (chars == ord('C')))
    gc = []
    for i in range(0, len(chars) - winsize + 1, stride):
        gc.append(np.mean(isgc[i:i+winsize]))
    return np.array(gc)

def compute_gc_penalty(seq, winsize, stride):
    gc = gc_content_sliding_window(seq, winsize, stride)
    return -(10 ** np.log2(np.abs(gc - 0.5) + 0.1)).sum()

class GCRatioFitness(ScoringFunction):

    single_submission = False

    name = 'gc'
    description = 'GC Ratio'
    priority = 50

    arguments = [
        ('weight', dict(
            type=float, default=3.0, metavar='WEIGHT',
            help='scoring weight for GC ratio (default: 3.0)')),
        ('window-size', dict(
            type=int, default=50, metavar='SIZE',
            help='size of window for GC content calculation (default: 50)')),
        ('stride', dict(
            type=int, default=5, metavar='STRIDE',
            help='size of stride for GC content calculation (default: 5)')),
    ]

    def __init__(self, weight, window_size, stride, _length_cds):
        num_windows = (_length_cds - window_size) // stride + 1
        num_windows = max(num_windows, 1)

        self.weight = weight / num_windows
        self.window_size = window_size
        self.stride = stride

    def score(self, seqs):
        gc_penalties = [compute_gc_penalty(seq, self.window_size, self.stride)
                        for seq in seqs]
        scores = [s * self.weight for s in gc_penalties]
        return {'gc_penalty': scores}, {'gc_penalty': gc_penalties}

    def evaluate_local(self, seq):
        gc = gc_content_sliding_window(seq, self.window_size, self.stride)
        centers = (
            np.arange(0, len(seq) - self.window_size + 1, self.stride) +
            self.window_size / 2)
        return {'gc': (centers, gc)}



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from . import ScoringFunction
from ..data import bicodon_usage_data
import numpy as np
from itertools import product

class BicodonAdaptationIndexFitness(ScoringFunction):

    single_submission = False

    name = 'bicodon'
    description = 'Codon Adaptation Index of Codon-Pairs'
    priority = 21

    requires = ['species']
    arguments = [
        ('weight', dict(
            type=float, default=1.0, metavar='WEIGHT',
            help='scoring weight for codon adaptation index of codon-pairs '
                 '(default: 1.0)'
        )),
    ]

    def __init__(self, weight, _length_cds, _species):
        self.weight = weight
        self.species = _species
        self.initialize_bicodon_scores()

    def initialize_bicodon_scores(self):
        if self.species not in bicodon_usage_data.bicodon_usage:
            raise ValueError(f'No bicodon usage data for species: {self.species}')

        bicodon_usage = (
            bicodon_usage_data.bicodon_usage[self.species].astype(np.float64))
        pairs = [''.join(seq) for seq in product('ACGU', repeat=6)]

        self.bicodon_scores = dict(zip(pairs, bicodon_usage))
        assert len(self.bicodon_scores) == 4096

    def score(self, seqs):
        if len(seqs[0]) < 6:
            return [0.0] * len(seqs)

        scores = self.bicodon_scores
        bcai = np.array([
            np.mean([scores[seq[i:i+6]] for i in range(0, len(seq) - 3, 3)])
            for seq in seqs])
        bcai_score = bcai * self.weight

        return {'bicodon': bcai_score}, {'bicodon': bcai}



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import json

execopts_to_remove = [
    'output', 'overwrite', 'quiet', 'seq_description', 'print_top_mutants',
    'protein']
iteroptions_aliases = {
    'n_iterations': 'iterations',
    'n_offsprings': 'offsprings',
    'n_survivors': 'survivors',
    'lineardesign_lambda': 'lineardesign',
}

def dump_to_preset(scoreopts, iteropts, execopts):
    data = {}

    data.update(iteropts._asdict())
    data.update({k: v for k, v in execopts._asdict().items()
                 if k not in execopts_to_remove})

    data['fitness'] = {}
    for modname, args in scoreopts.items():
        mod = data['fitness'][modname] = {}
        for argname, argval in args.items():
            if not argname.startswith('_'):
                mod[argname] = argval

    return json.dumps(data, indent=2)

def load_preset(data):
    return {
        iteroptions_aliases.get(k, k): v
        for k, v in json.loads(data).items()
    }



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import abc

class ScoringFunction(abc.ABC):

    name = 'noname'
    description = 'no description'

    # Command line help shows in ascending order of this priority
    priority = 99

    # If True, score() method is called with a list of folding predictions.
    uses_folding = False

    # Specifies the additional required arguments for the constructor
    requires = []

    # Command line arguments
    arguments = []

    # One letter flags for showing penalty metrics in the console.
    penalty_metric_flags = {}

    @classmethod
    def add_argument_parser(cls, parser):
        grp = parser.add_argument_group('Fitness - ' + cls.description)
        argprefix = '--{name}-'.format(name=cls.name.replace('_', '-'))
        argmap = []
        for argname, argopts in cls.arguments:
            grp.add_argument(argprefix + argname, **argopts)
            argmap.append((argprefix + argname, argname.replace('-', '_')))
        return argmap

    def __call__(self, seqs, *args, **kwargs):
        try:
            return self.score(seqs, *args, **kwargs)
        except KeyboardInterrupt:
            return None

    @abc.abstractmethod
    def score(self, seqs):
        raise NotImplementedError

def discover_scoring_functions(addon_paths):
    from . import __path__, __name__
    import pkgutil
    import importlib
    import os
    import sys

    funcs = {}

    def scan_module(mod):
        for objname in dir(mod):
            obj = getattr(mod, objname)
            if (obj is not ScoringFunction and type(obj) == abc.ABCMeta and
                    issubclass(obj, ScoringFunction)):
                funcs[obj.name] = obj

    for modinfo in pkgutil.iter_modules(__path__):
        modname = f'{__name__}.{modinfo.name}'
        mod = importlib.import_module(modname)
        scan_module(mod)

    loaded = set()
    for filepath in addon_paths:
        # It is tricky to load the addon module from each individual file because
        # multiprocessing requires the functions are importable using the module
        # path in sys.path. So we add the addon path to sys.path and load it using
        # the name.
        modname = os.path.splitext(os.path.basename(filepath))[0]
        abspath = os.path.abspath(filepath)
        if abspath in loaded:
            continue
        loaded.add(abspath)

        dirname = os.path.dirname(abspath)
        if dirname not in sys.path:
            sys.path.insert(0, os.path.dirname(abspath))
        mod = importlib.import_module(modname)
        scan_module(mod)

    return funcs



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from . import ScoringFunction

class LongStemFitness(ScoringFunction):

    name = 'longstem'
    description = 'RNA Folding (Long Stems)'
    priority = 43
    uses_folding = True

    arguments = [
        ('weight', dict(metavar='WEIGHT',
            type=float, default=100.0,
            help='penalty score for long stems (default: 100.0)')),
        ('threshold', dict(
            type=int, default=27, metavar='N',
            help='minimum length of stems to avoid (default: 27)')),
    ]

    penalty_metric_flags = {}

    def __init__(self, threshold, weight, _length_cds):
        self.threshold = threshold
        self.weight = -weight

        if weight != 0:
            self.penalty_metric_flags[self.name] = 'l'

    def score(self, seqs, foldings):
        metrics = []
        scores = []

        for fold in foldings:
            stems = fold['stems']
            longstems = sum(len(loc5) >= self.threshold for loc5, _ in stems)
            metrics.append(longstems)
            scores.append(longstems * self.weight)

        return {'longstem': scores}, {'longstem': metrics}

    def annotate_sequence(self, seq, folding):
        longstems = sum(len(loc5) >= self.threshold
                        for loc5, _ in folding['stems'])
        return {'longstems': longstems}



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from Bio.Data import CodonTable
from collections import namedtuple
import pandas as pd
import numpy as np
from . import lineardesign

STOP = '*'
MutationChoice = namedtuple('MutationChoice', ['pos', 'altcodon'])

class MutantGenerator:

    cdsseq = None
    initial_codons = None

    def __init__(self, cdsseq: str, random_state: np.random.RandomState,
                 codon_table: str='standard', is_protein: bool=False,
                 boost_loop_mutations: str=None):
        self.initialize_codon_table(codon_table)

        if is_protein:
            self.cdsseq = self.backtranslate(cdsseq)
        else:
            self.cdsseq = cdsseq
        if len(self.cdsseq) % 3 != 0:
            raise ValueError('Invalid CDS sequence length')

        self.rand = np.random if random_state is None else random_state

        if boost_loop_mutations is not None:
            tokens = boost_loop_mutations.split(':')
            (self.boost_loop_mutations_weight,
             self.boost_loop_mutations_start) = float(tokens[0]), int(tokens[1])
        else:
            self.boost_loop_mutations_weight = 0

        self.setup_choices()

    def initialize_codon_table(self, codon_table: str) -> None:
        table_var_name = f'{codon_table}_rna_table'
        if not hasattr(CodonTable, table_var_name):
            raise ValueError(f'Invalid codon table name: {codon_table}')

        self.codon_table = getattr(CodonTable, table_var_name)

        tbl = pd.DataFrame(
            list(self.codon_table.forward_table.items()) +
            [[stopcodon, STOP] for stopcodon in self.codon_table.stop_codons],
            columns=['codon', 'aa'])

        # Build synonymous codon lookup table
        self.synonymous_codons, self.aa2codons, self.codon2aa = {}, {}, {}
        for aa, codons in tbl.groupby('aa'):
            codons = set(codons['codon'])
            self.aa2codons[aa] = codons
            for c in codons:
                self.synonymous_codons[c] = sorted(codons - set([c]))
                self.codon2aa[c] = aa

    def backtranslate(self, proteinseq: str) -> str:
        return ''.join(next(iter(self.aa2codons[aa])) for aa in proteinseq)

    def translate(self, rnaseq: str) -> str:
        return ''.join(self.codon2aa[rnaseq[i:i+3]]
                       for i in range(0, len(rnaseq), 3))

    def setup_choices(self) -> None:
        choices = []
        initial_codons = []

        for i in range(len(self.cdsseq) // 3):
            codon = self.cdsseq[i*3:i*3+3]
            alternatives = len(self.synonymous_codons[codon])
            for alt in range(alternatives):
                choices.append(MutationChoice(i, alt))
            initial_codons.append(codon)

        self.choices = choices
        self.initial_codons = initial_codons

    def prepare_alternative_choices(self, left, right) -> None:
        return [choice for choice in self.choices if left <= choice.pos < right]

    def randomize_initial_codons(self) -> None:
        self.initial_codons[:] = [
            self.rand.choice([codon] + self.synonymous_codons[codon])
            for codon in self.initial_codons]

    def lineardesign_initial_codons(self, lmd, lddir, omitstart) -> None:
        prot = self.translate(self.cdsseq)

        prot_om = prot[omitstart:]
        res = lineardesign.run_lineardesign(lddir, prot_om, lmd)
        assert len(res['seq']) == len(res['str']) == len(prot_om) * 3

        rseq = res['seq']
        self.initial_codons[omitstart:] = [
            rseq[i*3:i*3+3] for i in range(len(prot_om))]

    def calc_probabilities(self, choices: list[MutationChoice],
                           folding: dict) -> np.ndarray:
        minimum_position = self.boost_loop_mutations_start
        loop_positions = set([i // 3 for i, code in enumerate(folding['folding'])
                              if code == '.'])

        weightmap = [1, self.boost_loop_mutations_weight]
        probs = [weightmap[mut.pos in loop_positions and mut.pos >= minimum_position]
                 for mut in choices]
        return np.array(probs) / sum(probs)

    def generate_mutant(self, codons: list[str], mutation_rate: float,
                        choices: list[MutationChoice]=None,
                        folding: dict=None) -> list[str]:
        child = codons[:]
        if choices is None:
            choices = self.choices

        # Draw number of mutations from binomial distribution
        n_mutations = self.rand.binomial(len(choices), mutation_rate)
        n_mutations = max(1, min(n_mutations, len(choices)))

        prob_dist = (
            self.calc_probabilities(choices, folding)
            if self.boost_loop_mutations_weight != 0 and folding is not None
            else None)

        # Select mutations
        mutation_choices = self.rand.choice(len(choices), n_mutations,
                                            replace=False, p=prob_dist)

        # Apply mutations
        for i in mutation_choices:
            mut = choices[i]
            child[mut.pos] = self.synonymous_codons[child[mut.pos]][mut.altcodon]

        return child

    def compute_expected_mutations(self, mutation_rate: float) -> float:
        return len(self.choices) * mutation_rate

    def compute_mutational_space(self) -> dict:
        log10_totalcases = np.sum(np.log10(list((map(len, self.choices)))))
        totalcases_mantissa = 10 ** (log10_totalcases % 1)
        totalcases = f'{totalcases_mantissa:.2f} x 10^{int(log10_totalcases)}'

        return {
            'singles': len(self.choices),
            'total': totalcases,
        }



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from . import ScoringFunction

ICODON_SPECIES_MAPPING = {
    'Homo sapiens': 'human',
    'Mus musculus': 'mouse',
    'Danio rerio': 'zebrafish',
}

class iCodonStabilityFitness(ScoringFunction):

    iCodon_initialized = False
    single_submission = False
    predfunc = None

    name = 'iCodon'
    description = 'iCodon'
    priority = 10

    requires = ['species']
    arguments = [
        ('weight', dict(
            type=float, default=1.0, metavar='WEIGHT',
            help='scoring weight for iCodon predicted stability (default: 1.0)')),
    ]

    def __init__(self, weight, _species, _length_cds):
        self.weight = weight
        if _species not in ICODON_SPECIES_MAPPING:
            raise ValueError(f"Unsupported species by iCodon: {_species}")
        self.species = ICODON_SPECIES_MAPPING[_species]
        self.length_cds = _length_cds

    def score(self, seqs):
        if not self.iCodon_initialized:
            import os
            os.environ['TZ'] = 'UTC' # dplyr requires this to run in singularity

            import rpy2.robjects.packages as rpackages
            rpackages.importr('iCodon')
            rpackages.importr('stringr')

            import rpy2.robjects as ro
            ro.r['options'](warn=-1)
            self.predfunc = ro.r['predict_stability'](self.species)

            self.iCodon_initialized = True

        # Remove duplicates since iCodon refuses prediction of sequence lists
        # containing duplicates
        query = list(set(seqs))
        results = self.predfunc(query)
        results = dict(zip(query, results))

        pred = [float(results[s]) for s in seqs]
        scores = [s * self.weight for s in pred]
        return {'pred_stability': scores}, {'pred_stability': pred}



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import sys
import re
import pylru
from tqdm import tqdm
from concurrent import futures
from collections import Counter
from .console import hbar_stars


class FoldEvaluator:

    def __init__(self, engine: str):
        self.engine = engine
        self.pat_find_loops = re.compile(r'\.{2,}')
        self.initialize()

    def initialize(self):
        if self.engine == 'vienna':
            try:
                import RNA
            except ImportError:
                raise ImportError('ViennaRNA module is not available. Try "'
                                  'pip install ViennaRNA" to install.')
            self._fold = RNA.fold
        elif self.engine == 'linearfold':
            try:
                import linearfold
            except ImportError:
                raise ImportError('LinearFold module is not available. Try "'
                                  'pip install linearfold-unofficial" to install.')
            self._fold = linearfold.fold
        else:
            raise ValueError(f'Unsupported RNA folding engine: {self.engine}')

    def __call__(self, seq):
        folding, mfe = self._fold(seq)
        stems = self.find_stems(folding)
        folding, stems = self.unfold_unstable_structure(folding, stems)
        loops = dict(Counter(map(len, self.pat_find_loops.findall(folding))))

        return {
            'folding': folding,
            'mfe': mfe,
            'stems': stems,
            'loops': loops,
        }

    @staticmethod
    def find_stems(structure):
        stack = []
        stemgroups = []

        for i, s in enumerate(structure):
            if s == '(':
                stack.append(i)
            elif s == ')':
                assert len(stack) >= 1
                peer = stack.pop()
                if (stemgroups and peer + 1 == stemgroups[-1][0][-1] and
                        i - 1 == stemgroups[-1][1][-1]):
                    stemgroups[-1][0].append(peer)
                    stemgroups[-1][1].append(i)
                else:
                    stemgroups.append(([peer], [i]))

        return stemgroups

    @staticmethod
    def unfold_unstable_structure(folding, stems):
        # TODO: This needs to be revised based on the thermodynamic model of RNA
        # folding later.
        lonepairs = [p for p in stems if len(p[0]) == 1]
        if not lonepairs:
            return folding, stems

        folding = list(folding)
        for p5, p3 in lonepairs:
            folding[p5[0]] = '.'
            folding[p3[0]] = '.'
        newstems = [p for p in stems if len(p[0]) > 1]

        return ''.join(folding), newstems


class SequenceEvaluator:

    folding_cache_size = 8192

    def __init__(self, scoring_funcs, scoreopts, execopts, mutantgen, species,
                 length_cds, quiet, printmsg):
        self.scoring_funcs = scoring_funcs
        self.scoreopts = scoreopts
        self.execopts = execopts

        self.length_cds = length_cds
        self.mutantgen = mutantgen
        self.species = species

        self.quiet = quiet
        self.printmsg = printmsg

        self.initialize()

    def initialize(self):
        self.foldeval = FoldEvaluator(self.execopts.folding_engine)
        self.folding_cache = pylru.lrucache(self.folding_cache_size)

        self.scorefuncs_nofolding = []
        self.scorefuncs_folding = []
        self.penalty_metric_flags = {}

        additional_opts = {
            '_length_cds': self.length_cds,
        }

        for funcname, cls in self.scoring_funcs.items():
            opts = self.scoreopts[funcname]
            if (('weight' in opts and opts['weight'] == 0) or
                ('off' in opts and opts['off'])):
                continue
            opts.update(additional_opts)
            for reqattr in cls.requires:
                opts['_' + reqattr] = getattr(self, reqattr)

            try:
                scorefunc_inst = cls(**opts)
            except EOFError:
                continue

            if cls.uses_folding:
                self.scorefuncs_folding.append(scorefunc_inst)
            else:
                self.scorefuncs_nofolding.append(scorefunc_inst)

            self.penalty_metric_flags.update(cls.penalty_metric_flags)

    def evaluate(self, seqs, executor):
        with SequenceEvaluationSession(self, seqs, executor) as sess:
            sess.evaluate()

            if not sess.errors:
                total_scores = [sum(s.values()) for s in sess.scores]
                return total_scores, sess.scores, sess.metrics, sess.foldings
            else:
                return None, None, None, None

    def get_folding(self, seq):
        if seq not in self.folding_cache:
            self.folding_cache[seq] = self.foldeval(seq)
        return self.folding_cache[seq]

    def prepare_evaluation_data(self, seq):
        folding = self.get_folding(seq)

        seqevals = {}
        seqevals['local-metrics'] = localmet = {}
        for fun in self.scorefuncs_nofolding + self.scorefuncs_folding:
            if hasattr(fun, 'evaluate_local'):
                localmet.update(fun.evaluate_local(seq))

            if hasattr(fun, 'annotate_sequence'):
                if fun.uses_folding:
                    seqevals.update(fun.annotate_sequence(seq, folding))
                else:
                    seqevals.update(fun.annotate_sequence(seq))

        return seqevals


class SequenceEvaluationSession:

    def __init__(self, evaluator: SequenceEvaluator, seqs: list[str],
                 executor: futures.Executor):
        self.seqs = seqs
        self.executor = executor

        self.scores = [{} for i in range(len(seqs))]
        self.metrics = [{} for i in range(len(seqs))]
        self.foldings = [None] * len(seqs)
        self.errors = []

        self.folding_cache = evaluator.folding_cache
        self.foldings_remaining = len(seqs)
        self.foldeval = evaluator.foldeval

        self.num_tasks = (
            len(evaluator.scorefuncs_folding) +
            len(evaluator.scorefuncs_nofolding) +
            len(seqs))

        self.pbar = None
        self.printmsg = evaluator.printmsg
        self.quiet = evaluator.quiet

        self.scorefuncs_folding = evaluator.scorefuncs_folding
        self.scorefuncs_nofolding = evaluator.scorefuncs_nofolding
    
    def __enter__(self):
        self.printmsg('')
        self.pbar = tqdm(total=self.num_tasks, disable=self.quiet,
                         file=sys.stderr, unit='task', desc='Scoring fitness')
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.pbar is not None:
            self.pbar.close()
        self.printmsg('')

    def evaluate(self) -> None:
        jobs = set()

        # Secondary structure prediction is the first set of tasks.
        for i, seq in enumerate(self.seqs):
            if self.errors: # skip remaining tasks on error
                continue

            if seq in self.folding_cache:
                self.foldings[i] = self.folding_cache[seq]
                self.foldings_remaining -= 1
                if self.pbar is not None:
                    self.pbar.update()
                continue

            future = self.executor.submit(self.foldeval, seq)
            future._seqidx = i
            future._type = 'folding'
            jobs.add(future)

        # Then, scoring functions that does not require folding are executed.
        for scorefunc in self.scorefuncs_nofolding:
            if self.errors:
                continue

            future = self.executor.submit(scorefunc, self.seqs)
            future._type = 'scoring'
            jobs.add(future)

        # Wait until all folding tasks are finished.
        while jobs and not self.errors and self.foldings_remaining > 0:
            done, jobs = futures.wait(jobs, timeout=0.1,
                                      return_when=futures.FIRST_COMPLETED)
            for future in done:
                if future._type == 'folding':
                    self.collect_folding(future)
                elif future._type == 'scoring':
                    self.collect_scores(future)

        # Scoring functions requiring folding are executed.
        for scorefunc in self.scorefuncs_folding:
            if self.errors:
                continue

            future = self.executor.submit(scorefunc, self.seqs,
                                          self.foldings)
            future._type = 'scoring'
            jobs.add(future)

        while jobs and not self.errors:
            done, jobs = futures.wait(jobs, timeout=0.1)
            for future in done:
                if future._type == 'folding':
                    self.collect_folding(future)
                elif future._type == 'scoring':
                    self.collect_scores(future)

    def collect_scores(self, future):
        try:
            ret = future.result()
            if ret is None:
                self.errors.append('KeyboardInterrupt')
                if self.pbar is not None:
                    self.pbar.close()
                self.pbar = None
                return
            scoreupdates, metricupdates = ret
        except Exception as exc:
            return self.handle_exception(exc)

        if self.pbar is not None:
            self.pbar.update()

        # Update scores
        for k, updates in scoreupdates.items():
            assert len(updates) == len(self.scores)
            for s, u in zip(self.scores, updates):
                s[k] = u

        # Update metrics
        for k, updates in metricupdates.items():
            assert len(updates) == len(self.metrics)
            for s, u in zip(self.metrics, updates):
                s[k] = u

    def collect_folding(self, future):
        try:
            folding = future.result()
            if folding is None:
                self.errors.append('KeyboardInterrupt')
                if self.pbar is not None:
                    self.pbar.close()
                self.pbar = None
                return
        except Exception as exc:
            return self.handle_exception(exc)
        i = future._seqidx
        self.foldings[i] = folding
        self.folding_cache[self.seqs[i]] = folding
        self.foldings_remaining -= 1

        if self.pbar is not None:
            self.pbar.update()

    def handle_exception(self, exc):
        self.printmsg(hbar_stars, force=True)
        self.printmsg('Error occurred in a scoring function:', force=True)

        import traceback
        import io
        errormsg = io.StringIO()
        traceback.print_exc(file=errormsg)
        self.printmsg(errormsg.getvalue(), force=True)

        self.printmsg(hbar_stars, force=True)
        self.printmsg(force=True)
        self.printmsg('Termination in progress. Waiting for running tasks '
                        'to finish before closing the program.', force=True)
        self.errors.append(exc.args)



#
# VaxPress
#
# Copyright 2023 Hyeshik Chang
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# “Software”), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from . import ScoringFunction
import pytrf

class TandemRepeatsFitness(ScoringFunction):

    name = 'repeats'
    description = 'Tandem Repeats'
    priority = 60

    arguments = [
        ('weight',
         dict(type=float, default=1.0, metavar='WEIGHT',
              help='scoring weight for tandem repeats (default: 1.0)')),
        ('min-repeats',
         dict(type=int, default=2, metavar='N',
              help='minimum number of repeats to be considered as a tandem '
                   'repeat (default: 2)')),
        ('min-length',
         dict(type=int, default=10, metavar='LENGTH',
              help='minimum length of repeats to be considered as a tandem '
                   'repeat (default: 10)')),
    ]

    penalty_metric_flags = {'repeat': 'r'}

    def __init__(self, weight, min_repeats, min_length, _length_cds):
        self.weight = weight / _length_cds * -1000
        self.min_repeats = min_repeats
        self.min_length = min_length

    def score(self, seqs):
        replengths = []
        for seq in seqs:
            repeats = pytrf.GTRFinder('name', seq, min_repeat=self.min_repeats,
                                    min_length=self.min_length)
            replengths.append(sum(r.length for r in repeats))

        repeat_score = [length * self.weight for length in replengths]

        metrics = {'repeat': replengths}
        scores = {'repeat': repeat_score}

        return scores, metrics
